// mesi.c: S.W. Ellingson, Virginia Tech, 2010 Nov 13
// ---
// REQUIRES: 
//   me.h
// ---
// This implements the function mesi() and is intended to be 
//   #include'd with another file that is calling it.  
// Used to inject commands into MCS/Scheduler's ms_exec process.  
//   (This is the client-side of a client-server connection)
// See end of this file for history.

#include "me.h"

/* error codes returned by mesi() */
#define MESI_ERR_OK         0 /* OK */
#define MESI_ERR_DEST       1 /* Didn't recognize "dest" argument */
#define MESI_ERR_CMD        2 /* Didn't recognize "cmd" argument */
#define MESI_ERR_CONNECT    4 /* connect() to scheduler (ms_exec) failed */
#define MESI_ERR_REJECTED   8 /* Scheduler (ms_exec) rejected the command */
#define MESI_ERR_DATA      16 /* "data" argument not valid for specified "cmd" */
#define MESI_ERR_FST_FOPEN 32 /* While processing DP_ FST, couldn't open specified .cf file */

int mesi( char *dest,  /* (input) DESTINATION.  (three-letter subsystem designator) */
                       /* see mcs.h for valid entries */
          char *cmd,   /* (input) TYPE. (three-letter command TYPE) */
                       /* see mcs.h for valid entries */
          char *data,  /* (input) DATA. (DATA field for given TYPE) -- depends on command */
          char *sdate, /* (input) Date when command should run.  Either MJD or "today".  If not specified, "today" is assumed. */
          char *stime, /* (input) Time when command should run.  Either MPM or "asap".  If not specified, "asap" is assumed. */
          long int *reference /* (output) REFERENCE returned by Scheduler (ms_exec); 0 if command was rejected */  
          ) {
// Returns an error code; see MESI_ERR_* #defines, above
// Concerning "cmd":
//   For subsystems NU#, SHL, ASP, and DR#, this is ASCII.
//     (The string provided will be exactly the string used for the DATA field)
// Concerning "data":
//   For dest="DP_", this will be a list of parameters that will get translated into a raw binary DATA field
//     For cmd="TBW": Args are TBW_BITS      {0|1}
//                             TBW_TRIG_TIME (samples from start of slot, uint32)
//                             TBW_SAMPLES   (samples, uint32)
//     For cmd="TBN": Args are TBN_FREQ      (Hz, float32) 
//                             TBN_BW        {1..7}
//                             TBN_GAIN      {0..30}
//                             sub_slot      {0..99}
//     For cmd="FST": Args are INDEX         ( -1, 0, or channel# (1-520) ) 
//                             cname         This is the name of a file, presumed to be located in 
//                                           MCS/Scheduler's "cfiles" 
//                                           directory, containing "sint16 COEFF_DATA[16][32]"
//     For cmd="BAM": Args are beam          1..NUM_BEAMS(4) (uint16 BEAM_ID) 
//                             dfile         This is the name of a file, presumed to be located in 
//                                           MCS/Scheduler's "dfiles" 
//                                           directory, containing "uint16 BEAM_DELAY[520]" 
//                             gfile         This is the name of a file, presumed to be located in 
//                                           MCS/Scheduler's "gfiles" 
//                                           directory, containing "sint16 BEAM_GAIN[260][2][2]"      
//                             sub_slot      {0..99}
//     For cmd="DRX": Args are beam          1..NUM_BEAMS(4)        (uint8 DRX_BEAM)
//                             tuning        1..NUM_TUNINGS(2)      (uint8 DRX_TUNING)
//                             freq          [Hz]                   (float32 DRX_FREQ)
//                             ebw  	     Bandwidth setting 1..7 (unit8 DRX_BW)
//                             gain          0..15                  (uint16 DRX_GAIN)
//                             subslot       0..99                  (uint8 sub_slot)
//   Remember to use quotes, since "data" includes spaces
// Note meaning of arguments is sensitive to position; so if you want <time> other than "asap", you have to
//   explicitly include a <date> parameter.
// Note that the <date>/<time> corresponds to the moment at which MCS/Scheduler (specifically, the ms_exec process)
//   will send the command to the specified subystem (specifically, to the corresponding ms_mcic process).  

  /*=================*/
  /*=== Variables ===*/
  /*=================*/

  int eResult = MESI_ERR_OK;  /* error code to be returned. */

  int sockfd;                 /* socket file discriptor */
  struct sockaddr_in address; /* for network sockets */

  struct LWA_cmd_struct c;    /* This structure defined in mcs.h */

  struct timeval tv;          /* from sys/time.h; included via me.h */
  struct timezone tz;

  short int i2u1, i2u2, i2u3;
  int i4u1, i4u2;
  float f41;                  /* assuming this is 32 bits */

  union {
    unsigned short int i;
    unsigned char b[2];
    } i2u;
  union {
    unsigned int i;
    unsigned char b[4];
    } i4u;
  union {
    float f;
    unsigned char b[4];
    } f4;

  unsigned char beam;
  unsigned char tuning;
  float freq;
  unsigned char ebw;
  unsigned short int gain;
  unsigned char subslot;
  
  //printf("[%d/%d] dest  = '%s'\n",ME_MESI,getpid(),dest);
  //printf("[%d/%d] cmd   = '%s'\n",ME_MESI,getpid(),cmd);
  //printf("[%d/%d] data  = '%s'\n",ME_MESI,getpid(),data);
  //printf("[%d/%d] sdate = '%s'\n",ME_MESI,getpid(),sdate);
  //printf("[%d/%d] stime = '%s'\n",ME_MESI,getpid(),stime);
  
  /*=== Fill in structure to send to ms_exec... ===*/

  /* figure out what subsystem */
  if ( !( c.sid = LWA_getsid( dest ) ) ) {
    printf("[%d/%d] FATAL: subsystem <%s> not recognized\n",ME_MESI,getpid(),dest);
    eResult += MESI_ERR_DEST;
    return eResult;    
    }

  /* Outbound value of ref doesn't matter. (Gets assigned by ms_exec.) */
  c.ref = 0;     

  /* figure out what command */
  if ( !( c.cid = LWA_getcmd( cmd ) ) ) {
    printf("[%d/%d] FATAL: cmd <%s> not recognized\n",ME_MESI,getpid(),cmd);
    eResult += MESI_ERR_CMD;
    return eResult; 
    }  
  //printf("[%s] c.cid=%d\n",ME,c.cid);

  ///* not implemented yet */
  //c.subslot = 0; 

  /* Not to be scheduled; just do as soon as time permits */
  c.bScheduled = 0;

  /* Outbound value sent to current time; won't matter because it's not */
  /* scheduled (see above) */
  gettimeofday( &c.tv, &tz );

  /* Outbound value doesn't matter, but gets changed by ms_exec */
  c.bAccept = 0; 

  /* Outbound value doesn't matter, but gets changed by ms_exec */
  c.eSummary = LWA_SIDSUM_NULL; 

  /* Outbound value doesn't matter, but gets changed by ms_mcic */
  c.eMIBerror = LWA_MIBERR_OK;

  /* Gets changed by ms_exec */
  strcpy(c.data,data); /* changed in reply */
  c.datalen = -1; /* assumed to be a string */

  /* For DP, c.data is raw binary, assembled from command line parameters. */
  /* the parameters are command-dependent */
  if (c.sid==LWA_SID_DP_) {
    
    switch (c.cid) {

       case LWA_CMD_PNG:
       case LWA_CMD_RPT:
       case LWA_CMD_SHT:
         break;
        
       case LWA_CMD_FST:
         // DATA field structure:
         // sint16 INDEX;
         // sint16 COEFF_DATA[16][32];

         // What we will send in this case is simply the string "data" as provided.
         // ms_mcic will convert this.
         strcpy(c.data,data); /* changed in reply */
         c.datalen = -1;      /* it's a string, for now */

         break;

       case LWA_CMD_BAM:
         // DATA field structure:
         // uint16 BEAM_ID;
         // uint16 BEAM_DELAY[520];
         // sint16 BEAM_GAIN[260][2][2];
         // uint8 sub_slot;

         // What we will send in this case is simply the string "data" as provided.
         // ms_mcic will convert this.
         strcpy(c.data,data); /* changed in reply */
         c.datalen = -1;      /* it's a string, for now */

         break;

       case LWA_CMD_DRX:
         // uint8 DRX_BEAM;
         // uint8 DRX_TUNING;
         // float32 DRX_FREQ;
         // unit8 DRX_BW;
         // uint16 DRX_GAIN;
         // uint8 sub_slot;

         // parse the input string into parameters
         sscanf(data,"%hhu %hhu %f %hhu %hu %hhu", &beam, &tuning, &freq, &ebw, &gain, &subslot);
         //printf("%f %1hhu %1hhu %1hhu %1hhu\n",freq,c.data[2],c.data[3],c.data[4],c.data[5]);

         // assemble into c.data:
         memcpy( &(c.data[0]), &beam,     1 );
         memcpy( &(c.data[1]), &tuning,   1 );

         /* flipping endian-ness of freq: */
         f4.f  = freq;  c.data[2]= f4.b[3]; c.data[3]= f4.b[2]; c.data[4]= f4.b[1]; c.data[5]= f4.b[0]; 

         memcpy( &(c.data[6]), &ebw,      1 );
         memcpy( &(c.data[7]), (&gain)+1, 1 ); /* flipping endian-ness of gain */
         memcpy( &(c.data[8]), (&gain)+0, 1 ); /* flipping endian-ness of gain */
         memcpy( &(c.data[9]), &subslot,  1 );

         f4.b[3] = c.data[2]; f4.b[2] = c.data[3]; f4.b[1] = c.data[4]; f4.b[0] = c.data[5]; freq = f4.f;  
         //printf("%f %1hhu %1hhu %1hhu %1hhu\n",freq,c.data[2],c.data[3],c.data[4],c.data[5]);

         c.datalen=10;

         break;

       case LWA_CMD_TBW:
         // DATA field structure:
         // uint8 TBW_BITS;
         // uint32 TBW_TRIG_TIME; 
         // uint32 TBW_SAMPLES;

         i2u1 = 0;
         i4u1 = 0;
         i4u2 = 0;
         sscanf(data,"%hu %u %u",&i2u1,&i4u1,&i4u2);
         //printf("[%d/%d] TBW args: TBW_BITS=%hu, TBW_TRIG_TIME=%u, TBW_SAMPLES=%u\n",ME_MESI,getpid(),i2u1,i4u1,i4u2);        
 
         i2u.i = i2u1;                     c.data[0]=i2u.b[0]; 
         i4u.i = i4u1; c.data[1]=i4u.b[3]; c.data[2]=i4u.b[2]; c.data[3]=i4u.b[1]; c.data[4]=i4u.b[0];
         i4u.i = i4u2; c.data[5]=i4u.b[3]; c.data[6]=i4u.b[2]; c.data[7]=i4u.b[1]; c.data[8]=i4u.b[0];

         c.datalen=9;

         break;

       case LWA_CMD_TBN:
         // DATA field structure:
         // float32 TBN_FREQ;
         // uint16 TBN_BW;
         // uint16 TBN_GAIN;
         // uint8 sub_slot;

         f41 = 0.0;
         i2u1 = 0;
         i2u2 = 0;
         i2u3 = 0;
         sscanf(data,"%f %hu %hu %hu",&f41,&i2u1,&i2u2,&i2u3);
         //printf("[%d/%d] TBN args TBN_FREQ=%f, TBN_BW=%hu, TBN_GAIN=%hu, sub_slot=%hu\n",ME_MESI,getpid(),f41,i2u1,i2u2,i2u3);

         f4.f  = f41;  c.data[0]= f4.b[3]; c.data[1]= f4.b[2]; c.data[2]= f4.b[1]; c.data[3]= f4.b[0];
         i2u.i = i2u1; c.data[4]=i2u.b[1]; c.data[5]=i2u.b[0]; 
         i2u.i = i2u2; c.data[6]=i2u.b[1]; c.data[7]=i2u.b[0]; 
         i2u.i = i2u3;                     c.data[8]=i2u.b[0]; 

         c.datalen=9;

         break;

       case LWA_CMD_INI:
         break;
 
       default:
         printf("[%d/%d] FATAL: cmd <%s> not recognized as valid for DP\n",ME_MESI,getpid(),cmd);
         eResult += MESI_ERR_CMD;
         return eResult;
         break;

       } /* switch (c.cid) */

    if (c.datalen > -1) { 
     char hex[256];
     LWA_raw2hex( c.data, hex, c.datalen );      
     //printf("[%d/%d] Outbound DATA field is: 0x%s (raw binary)\n",ME_MESI,getpid(),hex);  
    }

    } /* if (c.sid==LWA_SID_DP_) */

  /* create socket */
  sockfd = socket(
                  AF_INET,     /* domain; network sockets */
                  SOCK_STREAM, /* type (TCP-like) */
                  0);          /* protocol (normally 0) */
    
  /* network sockets */
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = inet_addr(LWA_IP_MSE);
  address.sin_port = htons(LWA_PORT_MSE);

  /* connect socket to server's socket */
  if (connect( sockfd, (struct sockaddr *) &address, sizeof(address))==-1) {
    eResult += MESI_ERR_CONNECT;
    return eResult;
    }

  write(sockfd, &c, sizeof(struct LWA_cmd_struct));
  read(sockfd,&c,sizeof(struct LWA_cmd_struct));

  //printf("[%d/%d] ref=%ld, bAccept=%d, eSummary=%d, data=<%s>\n",ME_MESI,getpid(),c.ref,c.bAccept,c.eSummary,c.data);

  close(sockfd); 

  if (!c.bAccept) {
      eResult += MESI_ERR_REJECTED;
      *reference = 0;
    } else {
      *reference = c.ref;
    }

  return eResult;

  } /* mesi() */

//==================================================================================
//=== HISTORY ======================================================================
//==================================================================================
// mesi.c: S.W. Ellingson, Virginia Tech, 2010 Nov 13
//   .1: Adding support for DP commands "FST", "BAM", and "DRX"
// mesi.c: S.W. Ellingson, Virginia Tech, 2010 Oct 15
//   .1: Initial version, splitting msei.c into this code and the rest into mesix.c
// --- NOTE NAME CHANGE ----
// msei.c: S.W. Ellingson, Virginia Tech, 2010 Oct 12
//   .1: Adding support for scheduled actions
// msei.c: S.W. Ellingson, Virginia Tech, 2010 Jun 10
//   .1: Adding support for MCS-DR.  Nothing to do besides update comments
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Nov 7
//   .1: Changed dimension of cmd[] and dest[] from 3 to 4
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Aug 25
//   .1: Adding support for DP's TBN, CLK, INI commands; got rid of "subslot"
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Aug 17
//   .1: Adding full support (binary argument) for DP's TBW command
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Aug 16
//   .1: Added "datalen" to LWA_cmd_struct, so this is set to "-1" to indicate "string"
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Aug 06
//   .1: Fixed bug causing DATA to be truncated to first space, even when quoted
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Aug 02
//   .1: Cleaning up console output
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Jul 26
//   .1: Added LWA_cmd_struct field eMIBerror (svn rev 10)
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Jul 23
//   .1: Adding "DATA" command line argument (svn rev 7)
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Jul 20
//   .1: Adding "time to execute" field to command struct
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Jul 17
//   .1: Cleaning up; esp. implementation of LWA_cmd_struct
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Jul 13
//   .1: 
// msei.c: S.W. Ellingson, Virginia Tech, 2009 Jul 10
//   .1: preliminary version (not even numbered); 3-letter char as argument 
//   .2: changing command line and transfer format

//==================================================================================
//=== BELOW THIS LINE IS SCRATCH ===================================================
//==================================================================================

